#!/usr/bin/env python
# Ubuntu Testing Automation Harness
# Copyright 2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Provisions and configures a phablet device for testing under UTAH."""

from __future__ import print_function

import argparse
import os
import socket
import subprocess
import sys
import threading
import time

from SimpleXMLRPCServer import SimpleXMLRPCServer

from utah.provisioning import touch
from utah import template

CLIENT_PROBES_DIR = '/tmp/utah-probes'


def phablet_tools_errors(f):
    """Decorating phablet-tools error management."""
    def _errors(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except subprocess.CalledProcessError:
            # phablet-flash prints a pretty good error, so we should just exit
            sys.exit(1)
    return _errors


@phablet_tools_errors
def _flash(args, device):
    cmd = ['phablet-flash']
    if args.serial:
        cmd.extend(['-s', args.serial])
    if args.revision:
        cmd.extend(['-r', args.revision])
    if args.uri:
        cmd.extend(['-u', args.uri])
    if args.device:
        cmd.extend(['-d', args.device])

    print('= flashing device...')
    subprocess.check_call(cmd)
    print()
    print('= waiting for install/boot...')
    device.wait_for_device()
    device.adb_root()
    device.wait_for_device()
    print()


@phablet_tools_errors
def _network(args):
    cmd = ['phablet-network-setup']
    if args.serial:
        cmd.extend(['-s', args.serial])
    if args.network_file:
        cmd.extend(['-n', args.network_file])

    print('= setting up network...')
    subprocess.check_call(cmd)
    time.sleep(5)  # give the device some time to aquire an IP
    print()


def _install_branch(device, branch):
    device.run_ubuntu('rm -rf /home/phablet/utah')
    device.run_ubuntu('bzr branch {} /home/phablet/utah'.format(branch))


def _install_utah(device, ppa, branch):
    print("= installing utah...")
    device.adb_root()
    device.wait_for_device()
    device.run_ubuntu('add-apt-repository -y {}'.format(ppa))
    device.run_ubuntu('apt-get update -qq')
    device.run_ubuntu('apt-get install -yq --force-yes utah-client')

    if branch:
        _install_branch(device, branch)


def _save_props(device, results_dir):
    props = device.run_ubuntu('getprop')
    props_file = os.path.join(results_dir, 'adb.props')
    with open(props_file, 'w') as f:
        f.write(props)


def _configure_autorun(device, event_name, env):
    autorun = '/usr/local/bin/utah-autorun.sh'
    buf = template.as_buff('utah-autorun.sh.jinja2', log_file='/tmp/utah.log')
    device.create_file_ubuntu(autorun, buf, '755')

    buf = 'phablet    ALL = NOPASSWD: ALL'
    device.create_file_ubuntu('/etc/sudoers.d/phablet', buf, '440')

    buf = [
        'start on {}'.format(event_name),
        'task',
        'exec {} >/tmp/utah.log 2>&1'.format(autorun),
    ]
    device.create_file_ubuntu('/etc/init/run_utah.conf', '\n'.join(buf))

    p = '/etc/utah/autorun'
    device.run_ubuntu('sh -c "[ -d {} ] || mkdir -p {}"'.format(p, p))

    if not env:
        env = ''
    buf = [
        '#!/bin/sh',
        'chmod 666 /dev/uinput',
        'export $(grep SESSION /home/phablet/.dbus-session)',
        'sudo -i -u phablet {} \\'.format(env),
        '  DBUS_SESSION_BUS_ADDRESS=$DBUS_SESSION_BUS_ADDRESS autopilot $*',
    ]
    device.create_file_ubuntu(
        '/usr/local/bin/utah-autopilot', '\n'.join(buf), '755')


def _get_phone_shell_cmds(args):
    if not args.no_shell:
        return '', ''

    sfile = '/etc/device-services'
    pre = [
        'if grep unity8 {} >/dev/null ; then '.format(sfile),
        '    echo "backing up device-services file"',
        '    cp {} {}.utah-bak'.format(sfile, sfile),
        'fi',
        'sed -i /unity8/d {}'.format(sfile),
        'restart ubuntu-touch-session',
    ]

    post = [
        '[ -f {}.utah-bak ] && mv {}.utah-bak {}'.format(sfile, sfile, sfile),
        'restart ubuntu-touch-session',
    ]

    return '\n'.join(pre), '\n'.join(post)


def _push_run_file(args, device, response_server=None):
    result = '/var/lib/utah/utah.yaml'
    pre_cmd, post_cmd = _get_phone_shell_cmds(args)
    cmds = [
        '#!/bin/sh',
        'set -x',
        '[ -d {} ] || rm -rf {}'.format(CLIENT_PROBES_DIR, CLIENT_PROBES_DIR),
        '[ -d {} ] || mkdir {}'.format(CLIENT_PROBES_DIR, CLIENT_PROBES_DIR),
        pre_cmd,
        '{} -d -o {} -r {} -p {}'.format(
            args.utah_cmd, result, args.runlist, CLIENT_PROBES_DIR),
        post_cmd,
    ]

    if response_server:
        cmds.insert(-1, 'sleep 2  #give time for usb disconnect script')
        cmds.append('if [ -f {} ] ; then'.format(result))
        cmds.append(('python -c "import xmlrpclib; '
                     's=xmlrpclib.ServerProxy(\'{}\', allow_none=True); '
                     's.utah_done()"'.format(response_server)))
        cmds.append('fi')
    device.create_file_ubuntu(
        '/etc/utah/autorun/01_run-utah', '\n'.join(cmds), '755')


def _run_utah_connected(args, device, event_name):
    """run utah and assumes we can safely use ADB throughout the test."""
    _push_run_file(args, device)

    try:
        with device.tail_syslog():
            device.run_ubuntu('initctl emit {}'.format(event_name))
    except touch.ADBShellError as e:
        # print the error but keep going so we can try and collect the logs
        # which may include a utah.yaml
        print(e.message)
    except KeyboardInterrupt as e:
        print('command interrupted, shutting down utah on target...')
        device.run_ubuntu('initctl stop {}'.format(event_name))
        raise e


def _public_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('8.8.8.8', 80))
    ip = s.getsockname()[0]
    s.close()
    return ip


def _run_utah_disconnected(args, device, event_name):
    """run utah with the USB connection to the device disabled."""

    def _utah_cb():
        print('client response received')

    # determine our IP and ephermal port for the client
    server = SimpleXMLRPCServer(('0.0.0.0', 0), allow_none=True)
    port = server.socket.getsockname()[1]

    _push_run_file(args, device, 'http://{}:{}'.format(_public_ip(), port))

    server.register_function(_utah_cb, 'utah_done')

    t = threading.Thread(target=server.handle_request)
    t.daemon = True
    t.start()
    print('= waiting for response on port {}...'.format(port))
    device.run_ubuntu('initctl emit --no-wait {}'.format(event_name))
    subprocess.check_call([args.disconnected, 'off'])
    t.join()
    subprocess.check_call([args.disconnected, 'on'])
    device.wait_for_device()


def _gather_results(args, device):
    dst = os.path.join(args.results_dir, 'utah.yaml')
    device.get_file_ubuntu('/var/lib/utah/utah.yaml', dst)

    dst = os.path.join(args.results_dir, 'utah.log')
    device.get_file_ubuntu('/tmp/utah.log', dst)

    device.get_file_ubuntu(CLIENT_PROBES_DIR, args.results_dir)

    for f in args.pull:
        device.get_file_ubuntu(f, args.results_dir)


def _run_utah(args, device):
    event_name = 'run_utah'
    _configure_autorun(device, event_name, args.env)

    if args.disconnected:
        _run_utah_disconnected(args, device, event_name)
    else:
        _run_utah_connected(args, device, event_name)
    _gather_results(args, device)


def _get_parser():
    parser = argparse.ArgumentParser(
        description='Provisions a device for UTAH testing.')
    parser.add_argument('-s', '--serial',
                        help='Android serial ID of device')
    parser.add_argument('--skip-install', action='store_true',
                        help='Skip running the phablet-flash install.')
    parser.add_argument('-r', '--revision',
                        help='series/revision to install')
    parser.add_argument('-u', '--uri',
                        help='Alternative download uri')
    parser.add_argument('--skip-network', action='store_true',
                        help='Skip setting up wifi settings on device.')
    parser.add_argument('-n', '--network-file',
                        help='''Specify network manager config file for wifi.
                        If none is specified we'll try and use the default wifi
                        configuration found on host system.''')
    parser.add_argument('-N', '--no-shell', action='store_true',
                        help='''Disable the unity shell. This is handy for
                        graphical tests like autopilot.''')
    parser.add_argument('--skip-utah', action='store_true',
                        help='Skip install the utah-client on device.')
    parser.add_argument('-b', '--branch',
                        help='Install UTAH client on device from branch.')
    parser.add_argument('-e', '--env',
                        help='Environment variables to set for utah-client.')
    parser.add_argument('--ppa', default='ppa:utah/stable',
                        help='''Specify an alternative PPA for utah.
                        default=%(default)s''')
    parser.add_argument('--results-dir', default='/tmp',
                        help='''Directory to store results in on the host.
                        default=%(default)s''')
    parser.add_argument('--pull', default=[], action='append',
                        help='''A file or directory to copy into the hosts's
                        results-dir after UTAH has been run. This option can
                        be specified multiple times.''')
    parser.add_argument('-l', '--runlist',
                        help='The utah runlist to execute')
    parser.add_argument('-d', '--disconnected',
                        help='''Uses the supplied script to disconnect USB
                        while UTAH is executed.''')
    parser.add_argument('--device',
                        help='The Android device type.')
    return parser


def _check_args(args, device):
    args.utah_cmd = 'utah'
    if args.branch:
        path = '/home/phablet/utah'
        args.utah_cmd = 'PYTHONPATH={} {}/client.py'.format(path, path)

    if args.env:
        args.utah_cmd = '{} {}'.format(args.env, args.utah_cmd)

    if not args.skip_install and not args.device and not device.detect_type():
        print('ERROR: --device option required for provisioning')
        sys.exit(1)

    if not os.path.exists(args.results_dir):
        print('ERROR: results-dir does not exist: {}'.format(args.results_dir))
        sys.exit(1)

    if not os.path.isdir(args.results_dir):
        print(
            'ERROR: results-dir not a directory: {}'.format(args.results_dir))
        sys.exit(1)


def main(argv=None):
    if argv:
        args = _get_parser().parse_args(argv)
    else:
        args = _get_parser().parse_args()

    device = touch.Device(print, args.serial)
    _check_args(args, device)

    if not args.skip_install:
        _flash(args, device)

    if not args.skip_network:
        _network(args)

    if not args.skip_utah:
        _install_utah(device, args.ppa, args.branch)

    _save_props(device, args.results_dir)

    if args.runlist:
        _run_utah(args, device)

if __name__ == "__main__":
    try:
        main()
    except touch.ADBShellError as e:
        print(e.long_message())
        sys.exit(1)
